#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct containergriddata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int szx,szy;
	int *minchildsizex;
	int *minchildsizey;
	int *optchildsizex;
	int *optchildsizey;
	int *asschildsizex;
	int *asschildsizey;
};

/* Special child routines, as children have position here.
 */
typedef struct ggiWidgetGridchild {
	struct ggiWidgetGridchild *next;
	ggi_widget_t              child;
	int x,y;
} *ggi_widget_gridchild_t;

static ggi_widget_gridchild_t ggiWidgetGridchildGetEntry(ggi_widget_t widget,int childnum) {

	int x;
	ggi_widget_gridchild_t current;
	ggi_widget_gridchild_t *rootptr=(ggi_widget_gridchild_t *)&widget->children;

	for(x=0,current=*rootptr;current;current=current->next,x++) {
		if (x==childnum||
		    (LAST_CHILD==childnum&&current->next==NULL))
		    	return current;
	}
	return NULL;
}

static ggi_widget_t ggiWidgetGridchildGet(ggi_widget_t widget,int childnum) {

	ggi_widget_gridchild_t current=ggiWidgetGridchildGetEntry(widget,childnum);

	if (!current) return NULL;
    	return current->child;
}


int ggiWidgetGridchildLink(ggi_widget_t self,int childnum,ggi_widget_t newchild,...) {

	int x,xp,yp;
	va_list args;

	ggi_widget_gridchild_t current,newcl;
	ggi_widget_gridchild_t *insptr;

	ggi_widget_gridchild_t *rootptr=(ggi_widget_gridchild_t *)&self->children;
	struct containergriddata *mine=self->priv;

	va_start(args,newchild);
	xp=va_arg(args,int);
	yp=va_arg(args,int);
	va_end(args);

	if (xp<0||xp>=mine->szx||
	    yp<0||yp>=mine->szy)
	    	return 1;

	current=*(insptr=rootptr);

	for(x=0;current;current=*(insptr=&current->next),x++) {
		if (x==childnum)
		    	break;
	}
	if (!current&&childnum!=LAST_CHILD)
		return 1;

	newcl=malloc(sizeof(*newcl));	
	if (!newcl)
		return 2;

	newcl->next=current;
	newcl->child=newchild;
	newcl->x=xp;
	newcl->y=yp;
	*insptr=newcl;
	newchild->parent=self;
	self->childchanged(self,newchild);
	return 0;
}

ggi_widget_t ggiWidgetGridchildUnlink(ggi_widget_t widget,int childnum) {

	int x;
	ggi_widget_gridchild_t current;
	ggi_widget_gridchild_t *insptr;
	ggi_widget_t rc=NULL;
	ggi_widget_gridchild_t *rootptr=(ggi_widget_gridchild_t *)&widget->children;

	current=*(insptr=rootptr);

	for(x=0;current;current=*(insptr=&current->next),x++) {
		if (x==childnum||
		    (LAST_CHILD==childnum&&current->next==NULL))
		    	break;
	}
	if (!current)
		return NULL;

	rc=current->child;
	current->child->parent=NULL;
	*insptr=current->next;
	free(current);
	widget->childchanged(widget,NULL);

	return rc;
}



static int containergrid_getsize(ggi_widget_t self) {

	struct containergriddata *mine=self->priv;
	ggi_widget_gridchild_t curchildlist;
	int x;

	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;

	for(x=0;x<mine->szx;x++) {
		mine->minchildsizex[x]=0;
		mine->optchildsizex[x]=0;
	}
	for(x=0;x<mine->szy;x++) {
		mine->minchildsizey[x]=0;
		mine->optchildsizey[x]=0;
	}

	for(x=0;(curchildlist=ggiWidgetGridchildGetEntry(self,x));x++) {

		if (mine->minchildsizex[curchildlist->x]<curchildlist->child->min.x)
			mine->minchildsizex[curchildlist->x]=curchildlist->child->min.x;
		if (mine->minchildsizey[curchildlist->y]<curchildlist->child->min.y)
			mine->minchildsizey[curchildlist->y]=curchildlist->child->min.y;

		if (mine->optchildsizex[curchildlist->x]<curchildlist->child->opt.x)
			mine->optchildsizex[curchildlist->x]=curchildlist->child->opt.x;
		if (mine->optchildsizey[curchildlist->y]<curchildlist->child->opt.y)
			mine->optchildsizey[curchildlist->y]=curchildlist->child->opt.y;

	}
	for(x=0;x<mine->szx;x++) {
		self->min.x+=mine->minchildsizex[x];
		self->opt.x+=mine->optchildsizex[x];
	}
	for(x=0;x<mine->szy;x++) {
		self->min.y+=mine->minchildsizey[x];
		self->opt.y+=mine->optchildsizey[x];
	}

	return 0;
}

static int ggiWidgetContainerGridRedraw(ggi_widget_t self,int flush) {

	struct containergriddata *mine=self->priv;
	ggi_widget_color_set *colset=ggiWidgetGetColorset(self);
	ggi_widget_gridchild_t curchildlist;
	int rc,x,y,z,xp,yp;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);

	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_empty));
	for(x=xp=0;x<mine->szx;xp+=mine->asschildsizex[x++]) {
		for(y=yp=0;y<mine->szy;yp+=mine->asschildsizey[y++]) {
			for(z=0;(curchildlist=ggiWidgetGridchildGetEntry(self,z));z++) {
				if (curchildlist->x==x&&curchildlist->y==y)
					goto is_handled;
			}
			ggiDrawBox(self->vis,self->pos.x+mine->drawpos.x+xp,self->pos.y+mine->drawpos.y+yp,
					mine->asschildsizex[x],mine->asschildsizey[y]);
			is_handled:;
		}
	}
	rc=ggiWidgetRedrawChildren(self);
	if (flush) {
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	}
	return rc;
}

static int ggiWidgetContainerGridVischanged(ggi_widget_t self) {

	struct containergriddata *mine=self->priv;
	ggi_widget_t curchild;
	ggi_widget_gridchild_t curchildentry;
	int sizeremainx,sizeremainy,inremain,x,dx,dy,y;

	if (self->vis==NULL) 
		return 0;

	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,0,0);

	if (mine->doopt) {
		sizeremainx=mine->drawsize.x - self->opt.x;
		inremain   =self->opt.x;
		for(x=0;x<mine->szx;x++) {
			mine->asschildsizex[x] =mine->optchildsizex[x];
			if (inremain<=0)dx=0;
			else		dx=sizeremainx*mine->optchildsizex[x]/inremain;
			mine->asschildsizex[x]+=dx;
			sizeremainx-=dx;
			inremain-=mine->optchildsizex[x];
		}
		if (sizeremainx!=0) {
			fprintf(stderr,"Grid layout doesn't fit branch opt x (%d).\n",sizeremainx);
		}
		sizeremainy=mine->drawsize.y - self->opt.y;
		inremain   =self->opt.y;
		for(y=0;y<mine->szy;y++) {
			mine->asschildsizey[y] =mine->optchildsizey[y];
			if (inremain<=0)dy=0;
			else		dy=sizeremainy*mine->optchildsizey[y]/inremain;
			mine->asschildsizey[y]+=dy;
			sizeremainy-=dy;
			inremain-=mine->optchildsizey[y];
		}
		if (sizeremainy!=0) {
			fprintf(stderr,"Grid layout doesn't fit branch opt y (%d).\n",sizeremainy);
		}
	} else {
		sizeremainx=mine->drawsize.x - self->min.x;
		inremain   =self->min.x;
		fprintf(stderr,"szrx=%d\n",sizeremainx);
		for(x=0;x<mine->szx;x++) {
			mine->asschildsizex[x] =mine->minchildsizex[x];
			if (inremain<=0) dx=0;
			else		 dx=sizeremainx*mine->minchildsizex[x]/inremain;
			mine->asschildsizex[x]+=dx;
			sizeremainx-=dx;
			inremain-=mine->minchildsizex[x];
		}
		if (sizeremainx!=0) {
			fprintf(stderr,"Grid Layout doesn't fit branch min x(%d).\n",sizeremainx);
		}
		sizeremainy=mine->drawsize.y - self->min.y;
		inremain   =self->min.y;
		for(y=0;y<mine->szy;y++) {
			mine->asschildsizey[y] =mine->minchildsizey[y];
			if (inremain<=0) dy=0;
			else		 dy=sizeremainy*mine->minchildsizey[y]/inremain;
			mine->asschildsizey[y]+=dy;
			sizeremainy-=dy;
			inremain-=mine->minchildsizey[y];
		}
		if (sizeremainy!=0) {
			fprintf(stderr,"Grid Layout doesn't fit branch min y(%d).\n",sizeremainy);
		}
	}

//	############### FIXME ###############

	for(x=0;(curchildentry=ggiWidgetGridchildGetEntry(self,x));x++) {
		int xx,yy,dx,dy;
		curchild=curchildentry->child;
		curchild->vis=self->vis;
		for(xx=dx=0;xx<curchildentry->x;xx++) {
			dx+=mine->asschildsizex[xx];
		}
		for(yy=dy=0;yy<curchildentry->y;yy++) {
			dy+=mine->asschildsizey[yy];
		}
		curchild->pos.x=self->pos.x+mine->drawpos.x+dx;
		curchild->pos.y=self->pos.y+mine->drawpos.y+dy;
		curchild->size.x=mine->asschildsizex[curchildentry->x];
		curchild->size.y=mine->asschildsizey[curchildentry->y];
		curchild->vischanged(curchild);
	}

	return 0;
}

static int ggiWidgetContainerGridChildchanged(ggi_widget_t self,ggi_widget_t child) {

	containergrid_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

ggi_widget_t ggiWidgetCreateContainerGrid(va_list ap) {

	ggi_widget_t newid;
	
	struct containergriddata *newbut;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newbut=malloc(sizeof(*newbut)))) {
		free(newid);
		return NULL;
	}
	newbut->szx=va_arg(ap,int);
	newbut->szy=va_arg(ap,int);
	newbut->minchildsizex=malloc(sizeof(newbut->minchildsizex[0])*newbut->szx);
	newbut->optchildsizex=malloc(sizeof(newbut->optchildsizex[0])*newbut->szx);
	newbut->asschildsizex=malloc(sizeof(newbut->asschildsizex[0])*newbut->szx);
	newbut->minchildsizey=malloc(sizeof(newbut->minchildsizex[0])*newbut->szy);
	newbut->optchildsizey=malloc(sizeof(newbut->optchildsizex[0])*newbut->szy);
	newbut->asschildsizey=malloc(sizeof(newbut->asschildsizex[0])*newbut->szy);
	newid->priv=newbut;

	newid->redraw      =ggiWidgetContainerGridRedraw;
	newid->vischanged  =ggiWidgetContainerGridVischanged;
	newid->childchanged=ggiWidgetContainerGridChildchanged;

	newid->getchild   =ggiWidgetGridchildGet;
	newid->linkchild  =ggiWidgetGridchildLink;
	newid->unlinkchild=ggiWidgetGridchildUnlink;

	containergrid_getsize(newid);

	return newid;
}
